import asyncio
import base64
import binascii
import hashlib
import hmac
from typing import Union
from xml.etree import ElementTree as ET

import bcrypt
from loguru import logger
from sqlalchemy import insert, select

from pyjabber import AppConfig
from pyjabber.db.database import DB
from pyjabber.db.model import Model
from pyjabber.features.SASL.Mechanism import MECHANISM
from pyjabber.features.SASL.utils import validate_cert, iq_register_result, not_authorized_response, success_response
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.stream.utils.Enums import Stage
from pyjabber.utils import ClarkNotation as CN
from pyjabber.utils.Exceptions import BadRequestException, InternalServerError, NotAuthorizerStreamNegotiationException


class SASL:
    """
        SASL Class.

        An instance of this class will manage the authentication from new connections,
        during the stream negotiators process.
    """
    __slots__ = ("_transport", "_parser", "_peer", "_from_claim", "_connection_manager",
                 "_loop", "_semaphore", "_pool_executor", "_handlers")

    def __init__(self, transport, parser, peer, from_claim = None):
        self._transport = transport
        self._parser = parser
        self._peer = peer
        self._from_claim = from_claim

        self._connection_manager = ConnectionManager()
        self._loop = asyncio.get_running_loop()

        self._semaphore = AppConfig.app_config.semaphore
        self._pool_executor = AppConfig.app_config.process_pool_exe

        self._handlers = {
            "iq": self.handle_iq,
            "auth": self.handle_auth
        }

    async def feed(self, element: ET.Element) -> Union[Stage, None]:
        _, tag = CN.break_down(element.tag)
        return await self._handlers[tag](element)

    async def handle_iq(self, element: ET.Element) -> None:
        query = element.find("{jabber:iq:register}query")
        if query is None:
            raise BadRequestException()

        if element.attrib.get("type") == IQ.TYPE.SET.value:
            new_jid = query.find("{jabber:iq:register}username")
            if new_jid is None:
                raise BadRequestException()

            new_jid = new_jid.text

            async with await DB.connection_async() as con:
                query_db = select(Model.Credentials).where(
                    Model.Credentials.c.jid == new_jid
                )
                credentials = await con.execute(query_db)
                credentials = credentials.fetchone()

            if credentials:
                self._transport.write(
                    SE.conflict_error(element.attrib.get("id"))
                )
            else:
                pwd = query.find("{jabber:iq:register}password").text
                await self._store_hash_task(pwd, new_jid)
                self._transport.write(iq_register_result(element.attrib["id"]))

        elif element.attrib.get("type") == IQ.TYPE.GET.value:
            iq = IQ(
                type_=IQ.TYPE.RESULT,
                id_=element.attrib["id"] if "id" in element.attrib.keys() else None,
            )
            query = ET.Element('{jabber:iq:register}query')
            ET.SubElement(query, "username")
            ET.SubElement(query, "password")
            iq.append(query)

            self._transport.write(ET.tostring(iq))

    async def handle_auth(self, element: ET.Element) -> Union[Stage, None]:
        mechanism = element.attrib.get("mechanism")

        if mechanism == MECHANISM.EXTERNAL.value:   # S2S SASL process
            if not self._from_claim:
                raise BadRequestException()

            cert = self._connection_manager.get_connection_ssl_certificate(self._peer)
            if not cert:
                logger.error(f"Error retrieving TLS cert from {self._peer}")
                self._transport.write(not_authorized_response())
                return

            if not validate_cert(self._from_claim, cert):
                logger.error(
                    f"Host claim cannot be verified with presented cert. Verify host used on stream or cert: {self._peer}"
                )
                self._transport.write(not_authorized_response())
                return

            self._transport.write(success_response())
            self._parser.reset_stack()
            return Stage.AUTH

        try:    # C2S SASL process
            data = base64.b64decode(element.text).split("\x00".encode())
            jid = data[1].decode()
            pwd = data[2].decode()

            async with await DB.connection_async() as con:
                query = select(Model.Credentials.c.hash_pwd).where(Model.Credentials.c.jid == jid)
                hashed_pwd = await con.execute(query)
                hashed_pwd = hashed_pwd.fetchone()

            if not hashed_pwd:
                self._transport.write(SE.not_authorized_sasl())
                self._connection_manager.close(self._peer)

            authorized = await self._verify_password_async(
                stored_password=hashed_pwd[0],
                provided_password=pwd
            )
            if authorized:
                self._connection_manager.set_jid(self._peer, JID(user=jid, domain=AppConfig.app_config.host))
                self._transport.write(success_response())
                self._parser.reset_stack()
                self._transport.resume_reading()
                return Stage.AUTH
            else:
                self._transport.write(SE.not_authorized_sasl())
                self._connection_manager.close(self._peer)

        except Exception as e:
            logger.error(f"Exception during auth process for {element.attrib.get('from')}: {e}")
            self._transport.write(SE.not_authorized_sasl())

    ###############################################################################################
    async def _store_hash_task(self, password: str, jid_str: str):
        """Handles the full user registration process by hashing the password and storing credentials.

            This coroutine executes the CPU-intensive password hashing and the I/O-bound
            database insertion, ensuring neither operation blocks the main event loop.
            It concludes the registration by sending an XMPP success response to the client.

            Args:
                password: The user's plain text password to be hashed and stored.
                jid_str: The JID of the user being registered.
        """
        hashed_pwd = await self._hash_scram_async(
            password=password,
            iterations=100000,
            salt=bcrypt.gensalt(rounds=8) if AppConfig.app_config.database_in_memory else bcrypt.gensalt()
        )

        async with await DB.connection_async() as con:
            async with con.begin():
                query_insert = insert(Model.Credentials).values({"jid": jid_str, "hash_pwd": hashed_pwd})
                await con.execute(query_insert)

    async def _verify_password_async(self, stored_password: str, provided_password: str) -> bool:
        """
        Verifies a password against the stored string.

        Args:
            stored_password: The full string from the database (e.g., "sha256$100000$salt_hex$hash_hex").
            provided_password: The password entered by the user.

        Returns:
            bool: True if the passwords match, False otherwise.
        """
        try:
            _, iter_str, salt_hex, hash_hex = stored_password.split('$')

            iterations = int(iter_str)
            salt = binascii.unhexlify(salt_hex)

            new_hash = await self._hash_scram_async(provided_password, salt, iterations)
            _, _, _, new_hash = new_hash.split("$")

            return hmac.compare_digest(new_hash, hash_hex)

        except (ValueError, IndexError, binascii.Error):
            return False

    async def _hash_scram_async(self, password: str, salt: bytes, iterations: int) -> str:
        """Calculates the PBKDF2 derived key (used in SCRAM) asynchronously.
            This function offloads the CPU-intensive PBKDF2-HMAC operation to a
            process pool executor to prevent blocking the asyncio event loop.
            It then encodes the result into a single secure string for storage.

            Args:
                password: The user's plain text password.
                salt: A unique, random byte string for this user.
                iterations: The number of iterations (cost) for the PBKDF2 calculation.

            Returns:
                str: A formatted string containing the algorithm, iterations, salt (in hex),
                     and the resulting hash (in hex), separated by '$' (e.g., "sha256$100000$salt_hex$hash_hex").
                     This format is ideal for securely storing credentials in a single
                     database column.
        """
        hashed_pwd = None
        async with self._semaphore:
            hashed_pwd = await self._loop.run_in_executor(
                self._pool_executor,
                hashlib.pbkdf2_hmac,
                'sha256',
                password.encode(),
                salt,
                iterations
            )

        salt_hex = binascii.hexlify(salt).decode('ascii')
        hash_hex = binascii.hexlify(hashed_pwd).decode('ascii')

        return f"sha256${iterations}${salt_hex}${hash_hex}"



import re
import xml.etree.ElementTree as ET
from typing import Dict

from pyjabber import AppConfig

# Plugins
from pyjabber.plugins.roster.Roster import Roster
from pyjabber.plugins.xep_0009.xep_0009 import RPC
from pyjabber.plugins.xep_0030.xep_0030 import Disco
from pyjabber.plugins.xep_0060.xep_0060 import PubSub
from pyjabber.plugins.xep_0199.xep_0199 import Ping
from pyjabber.plugins.xep_0363.xep_0363 import HTTPFieldUpload
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN


class PluginManager:
    __slots__ = ('_jid', '_plugins')

    def __init__(self, jid: JID) -> None:
        self._jid = jid

        self._plugins: Dict[str, object] = {
            'jabber:iq:roster': Roster(),
            'urn:xmpp:ping': Ping,
            'urn:xmpp:http:upload:0': HTTPFieldUpload(),
            'jabber:iq:rpc': RPC()
        }

        if any(p.startswith('http://jabber.org/protocol/disco') for p in AppConfig.app_config.plugins):
            self._plugins['http://jabber.org/protocol/disco*'] = Disco()
        if any(p.startswith('http://jabber.org/protocol/pubsub') for p in AppConfig.app_config.plugins):
            self._plugins['http://jabber.org/protocol/pubsub*'] = PubSub()

    async def feed(self, element: ET.Element):
        try:
            child = element[0]
        except IndexError:
            if element.attrib["type"] == "result":
                return None # Safe return. Nothing to process
            else:
                return SE.bad_request()

        ns, tag = CN.break_down(child.tag)

        try:
            ns = list(filter(lambda regex: re.search(regex, ns), list(self._plugins.keys())))[-1]
            return await self._plugins[ns].feed(self._jid, element)
        except (KeyError, IndexError):
            return SE.feature_not_implemented(tag, ns)



import asyncio
from typing import Union
from xml.etree import ElementTree as ET

from loguru import logger

from pyjabber import AppConfig
from pyjabber.network.utils.TransportProxy import TransportProxy
from pyjabber.queues.NewConnection import NewConnectionWrapper
from pyjabber.queues.QueueManager import get_queue, QueueName
from pyjabber.stream.negotiators.StreamNegotiator import Signal, Stage, StreamNegotiator


class StreamServerNegotiator(StreamNegotiator):
    def __init__(self, transport, protocol, parser, handler) -> None:
        super().__init__(transport, protocol, parser, handler)

        self._connection_queue = get_queue(QueueName.CONNECTIONS)

        self._stages_handlers = {
            Stage.CONNECTED: self._handle_init,
            Stage.OPENED: self._handle_tls,
            Stage.SSL: self._handle_init_ssl,
            Stage.SASL: self._handle_ssl,
            Stage.AUTH: self._handle_init_resource_bind,
            Stage.BIND: self._handle_resource_bind
        }

        self._stages_flags = {
            "starttls": False,
            "starttls_handshake": False,
            "auth": False,
        }

    async def handle_open_stream(self, elem: ET.Element = None) -> Union[Signal, None]:
        if elem.tag == "{http://etherx.jabber.org/streams}features":
            children = [child.tag for child in elem]
            if ("{urn:ietf:params:xml:ns:xmpp-tls}starttls" in children
                and self._stages_flags["starttls"] is False):
                    self._transport.write("<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>".encode())
                    self._stages_flags["starttls"] = True

            elif ("{urn:ietf:params:xml:ns:xmpp-sasl}mechanisms" in children
                  and self._stages_flags["starttls"]
                  and self._stages_flags["starttls_handshake"]):

                mechanisms_offered = [m for m in elem.find("{urn:ietf:params:xml:ns:xmpp-sasl}mechanisms")]
                if any('EXTERNAL' == m.text for m in mechanisms_offered):
                    self._transport.write(
                        "<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' mechanism='EXTERNAL'>=</auth>".encode())

        elif (elem.tag == "{urn:ietf:params:xml:ns:xmpp-tls}proceed"
              and self._stages_flags["starttls"]
              and self._stages_flags["starttls_handshake"] is False):

                await self.handle_tls_server()
                self._parser.reset_stack()
                self._stages_flags["starttls_handshake"] = True

        elif elem.tag == "{urn:ietf:params:xml:ns:xmpp-sasl}success":
            self._stages_flags["auth"] = True

        if all(self._stages_flags.values()):
            self._parser.reset_stack()
            return Signal.DONE
        else:
            return None

    async def handle_tls_server(self):
        transport_proxy_is_used = isinstance(self._transport, TransportProxy)
        if transport_proxy_is_used:
            original_transport = self._transport.original_transport
        else:
            original_transport = self._transport

        try:
            loop = asyncio.get_running_loop()
            new_transport = await loop.start_tls(
                transport=original_transport,
                protocol=self._protocol,
                sslcontext=AppConfig.app_config.ssl_context_s2s,
            )

            if transport_proxy_is_used:
                new_transport = TransportProxy(new_transport, self._peer)

            self._transport = new_transport
            self._protocol.transport = new_transport
            self._parser.transport = new_transport
            self._handler.transport = new_transport
            self._connection_manager.update_transport_server(new_transport, self._peer)

            logger.debug(f"Done TLS for <{self._peer}>")
        except ConnectionResetError as e:
            logger.error(f"Error during TLS upgrade with <{self._peer}> Reason: {e}")
            self._connection_manager.close(self._peer)
            return Signal.FORCE_CLOSE



# https://sabre.io/xml/clark-notation/
import re
from typing import Tuple
from xml.etree.ElementTree import Element


def break_down(tag: str):
    """
    Return the namespace and tag separated in a 2-tuple
    :return: (namespace, tag)
    """
    namespace = tag.split("}")[0].replace("{", "")
    tag = tag.split("}")[1]
    return namespace, tag


def clark_from_tuple(tuple: Tuple[str, str]):
    """
    Generate a clark notaion from a tuple in the format (namespace, tag).
    It's also possible to pass a tuple form of only a tag element (tag)
    """
    if tuple[0] is None:
        return f"{tuple[1]}"
    return f"{{{tuple[0]}}}{tuple[1]}"

def update_namespace(ns: str, element: Element):
    _, tag = break_down(element.tag)
    element.tag = f'{{{ns}}}{tag}'

    for child in element:
        update_namespace(ns, child)

def is_clark(tag: str) -> bool:
    """
    Regex to check if a string is in the clark notation format
    """
    return re.match(
        "^\\{[a-zA-Z0-9:/#?=&;.]+?\\}[a-zA-Z_][a-zA-Z0-9_]*$",
        tag) is not None



from uuid import uuid4

import pytest

from pyjabber.stream.JID import JID


def test_standard_jid():
    jid = JID(jid='demotest@host')

    assert jid.user == 'demotest'
    assert jid.domain == 'host'
    assert jid.resource is None


def test_user_and_domain_jid():
    jid = JID(user='demotest', domain='host')

    assert jid.user == 'demotest'
    assert jid.domain == 'host'
    assert jid.resource is None


def test_user_and_domain_and_resource_jid():
    resource = str(uuid4())
    jid = JID(user='demotest', domain='host', resource=resource)

    assert jid.user == 'demotest'
    assert jid.domain == 'host'
    assert jid.resource == resource


def test_to_string_jid():
    jid_string = f'demo@host/{str(uuid4())}'
    jid = JID(jid_string)

    assert jid_string == str(jid)


def test_eq():
    jid_1 = JID("demo@host")
    jid_2 = JID(user="demo", domain="host")

    assert (jid_1 == jid_2) is True

    jid_3 = JID("demo@host/res1")
    jid_4 = JID(user="demo", domain="host", resource="res1")

    assert (jid_3 == jid_4) is True

    jid_5 = JID("demo@host")
    jid_6 = JID("demo@host/res1")

    assert (jid_5 == jid_6) == False


def test_bad_arguments():
    with pytest.raises(ValueError):
        JID(jid='demo')
    with pytest.raises(ValueError):
        JID(jid='demo/test')
    with pytest.raises(ValueError):
        JID(jid='demo')
    with pytest.raises(ValueError):
        JID(domain='host', resource='1234')
    with pytest.raises(ValueError):
        JID(user='host', resource='1234')



from xml.etree import ElementTree as ET
from uuid import uuid4

from pyjabber import AppConfig


def validate_cert(from_claim, cert):
    peer_cert = cert.getpeercert()
    cert_names = set()

    subject = peer_cert.get("subject", [])
    for attr in subject:
        for (key, value) in attr:
            if key == "commonName":
                cert_names.add(value)

    for typ, value in peer_cert.get("subjectAltName", []):
        if typ == "DNS":
            cert_names.add(value)

    return from_claim in cert_names

def iq_register_result(iq_id: str) -> bytes:
    iq = ET.Element(
        "iq",
        attrib={
            "type": "result",
            "id": iq_id or str(uuid4()),
            "from": AppConfig.app_config.host})
    return ET.tostring(iq)

def not_authorized_response() -> bytes:
    return b"<failure xmlns='urn:ietf:params:xml:ns:xmpp-sasl'><not-authorized/></failure></stream>"

def success_response() -> bytes:
    return b"<success xmlns='urn:ietf:params:xml:ns:xmpp-sasl'/>"



from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class NewConnectionWrapper:
    """
    Represents a new connection made.
    It can be from a client or a server.
    """
    value: str
    client: bool = True



from enum import Enum


class PresenceShow(Enum):
    EXTENDED_AWAY = "xa"
    AWAY = "away"
    CHAT = "chat"
    DND = "dnd"
    NONE = "none"


class PresenceType(Enum):
    AVAILABLE = "available"
    UNAVAILABLE = "unavailable"



import asyncio
import xml.etree.ElementTree as ET
from typing import List, Tuple, Union
from uuid import uuid4
from xml.etree.ElementTree import Element

from sqlalchemy import delete, select

from pyjabber import AppConfig
from pyjabber.db.database import DB
from pyjabber.db.model import Model
from pyjabber.features.presence.Enums import PresenceShow, PresenceType
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.plugins.roster.Roster import Roster
from pyjabber.queues.NewConnection import NewConnectionWrapper
from pyjabber.queues.QueueManager import get_queue, QueueName
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class Presence(metaclass=Singleton):
    __slots__ = ('_handlers', '_connections', '_roster', '_pending', '_connection_queue', '_internal_queue', '_online_status')

    def __init__(self) -> None:
        self._handlers = {
            "subscribe": self._handle_subscribe,
            "subscribed": self._handle_subscribed,
            "unsubscribed": self._handle_unsubscribed,
            "unavailable": self._handle_global_presence,
            "INTERNAL": self._handle_lost_connection
        }

        self._connections = ConnectionManager()
        self._connection_queue = get_queue(QueueName.CONNECTIONS)
        self._internal_queue = asyncio.Queue()

        self._roster = Roster()
        self._pending = {}

        self._online_status = {}

    async def get_all_pending_presence(self):
        async with await DB.connection_async() as con:
            query = select(Model.PendingSubs.c.jid, Model.PendingSubs.c.item)
            res = await con.execute(query)
            res = res.fetchall()

        for jid_from, jid_to, item in res:
            if jid_to not in self._pending:
                self._pending[jid_to] = [item]
            else:
                self._pending[jid_to].append(item)

    async def update_pending_presence(self):
        pass

    async def delete_pending_presence(self, jid: str):
        async with await DB.connection_async() as con:
            query = delete(Model.PendingSubs).where(Model.PendingSubs.c.jid == jid)
            await con.execute(query)
            if not AppConfig.app_config.database_in_memory:
                await con.commit()

        self._pending[jid].pop()

    def priority_by_jid(self, jid: JID):
        try:
            return self._online_status[jid.bare()]
        except KeyError:
            return []

    def most_priority(self, jid: JID) -> List[Tuple[str, PresenceType, Union[str, None], Union[str, None], Union[str, None]]]:
        priority = self.priority_by_jid(jid)

        values = [item[-1] for item in priority if item[-1] is not None]
        max_value = max(values) if values else None
        return [item for item in priority if item[-1] == max_value]

    async def put(self, value):
        await self._internal_queue.put(value)

    def put_nowait(self, value):
        self._internal_queue.put_nowait(value)

    async def feed(self, jid: JID, element: Element):
        if "to" not in element.attrib:
            return await self._handle_global_presence(jid, element)
        else:
            if "type" in element.attrib:
                return await self._handlers[element.get("type")](jid, element)
            else:
                return await self._handle_directed_presence(jid, element)

    async def _present_in_online_list(self, jid: JID):
        if jid.bare() not in self._online_status:
            return None, None

        present = [(i,p) for i, p in enumerate(self._online_status[jid.bare()]) if p[0] == jid.resource]
        if present:
            return present[0]
        return None, None


    def _handle_lost_connection(self, jid: Union[JID | str], element: Element):
        if isinstance(jid, JID):
            if jid.bare() not in self._online_status:
                return None

            index, present = self._present_in_online_list(jid)
            if present:
                if self._online_status[jid.bare()][index][1] == PresenceType.UNAVAILABLE:
                    self._online_status[jid.bare()].pop(index)
                    if len(self._online_status[jid.bare()]) == 0:
                        self._online_status.pop(jid.bare())
                    return None

            for contact in self._roster.roster_by_jid(jid):
                item = ET.fromstring(contact.get("item"))
                if item.attrib.get("subscription") not in ["from", "both"]:
                    continue

                contact_jid = item.attrib.get("jid")

                if len(contact_jid.split("@")) < 2:
                    contact_jid = JID(contact_jid + f"@{AppConfig.app_config.host}")
                else:
                    contact_jid = JID(contact_jid)

                if contact_jid.bare() in self._online_status:
                    for user_connected in [i for i in self._online_status[contact_jid.bare()] if
                                           i[1] == PresenceType.AVAILABLE]:
                        resource = user_connected[0]
                        client = self._connections.get_transport(
                            JID(user=contact_jid.user, domain=contact_jid.domain, resource=resource))[0]

                        presence = ET.Element(
                            "presence",
                            attrib={
                                "from": str(jid),
                                "to": client.jid.bare(),
                                "type": PresenceType.UNAVAILABLE.value
                            }
                        )
                        client.transport.write(ET.tostring(presence))

            for client in self._connections.get_transport(JID(jid.bare())):
                presence = ET.Element(
                    "presence",
                    attrib={
                        "from": str(jid),
                        "to": jid.bare(),
                        "type": PresenceType.UNAVAILABLE.value
                    }
                )
                client.transport.write(ET.tostring(presence))
        else:
            pass    #TODO: presence for remote server lost

    async def _handle_lost_connection_server(self, host: str, element: Element):
        pass

    async def _handle_global_presence(self, jid: JID, element: ET.Element):
        if jid.bare() not in self._online_status:
            self._online_status[jid.bare()] = []

        show = next((c.text for c in element if c.tag == '{jabber:client}show' and c.text in [i.value for i in PresenceShow]), None)
        status = next((c.text for c in element if c.tag == '{jabber:client}status'), None)
        priority = next((c.text for c in element if c.tag == '{jabber:client}priority'), None)

        index, present = await self._present_in_online_list(jid)
        if element.attrib.get("type") == PresenceType.UNAVAILABLE.value:
            if present:
                self._online_status[jid.bare()][index] = (
                    jid.resource, PresenceType.UNAVAILABLE, show, status, priority
                )
            else:
                self._online_status[jid.bare()].append(
                    (jid.resource, PresenceType.UNAVAILABLE, show, status, priority)
                )

            self._connections.online(jid, False)
        else:
            if present:
                previous_status = self._online_status[jid.bare()][index][1]
                self._online_status[jid.bare()][index] = (jid.resource, PresenceType.AVAILABLE, show, status, priority)
                if previous_status == PresenceType.UNAVAILABLE:
                    await self._connection_queue.put(
                        NewConnectionWrapper(jid)
                        # ('CONNECTION', jid)
                    )
            else:
                self._online_status[jid.bare()].append(
                    (jid.resource, PresenceType.AVAILABLE, show, status, priority)
                )
                await self._connection_queue.put(
                    NewConnectionWrapper(jid)
                    # ('CONNECTION', jid)
                )

            self._connections.online(jid)

        for contact in self._roster.roster_by_jid(jid):
            item = ET.fromstring(contact.get("item"))
            if item.attrib.get("subscription") not in ["from", "both"]:
                continue

            contact_jid = item.attrib.get("jid")

            if len(contact_jid.split("@")) < 2:
                contact_jid = JID(contact_jid + f"@{AppConfig.app_config.host}")
            else:
                contact_jid = JID(contact_jid)

            if contact_jid.bare() in self._online_status:
                for user_connected in [i for i in self._online_status[contact_jid.bare()] if i[1] == PresenceType.AVAILABLE]:
                    resource = user_connected[0]
                    online = self._connections.get_transport_online(JID(user=contact_jid.user, domain=contact_jid.domain, resource=resource))
                    if online:
                        online = online.pop()
                        presence = ET.Element(
                            "presence",
                            attrib={
                                "from": str(jid),
                                "to": online.jid.bare(),
                            }
                        )

                        if element.attrib.get("type") == PresenceType.UNAVAILABLE.value:
                            presence.attrib["type"] = PresenceType.UNAVAILABLE.value

                        if show:
                            show_et = ET.SubElement(presence, "show")
                            show_et.text = show

                        if status:
                            status_et = ET.SubElement(presence, "status")
                            status_et.text = status

                        if priority:
                            priority_et = ET.SubElement(presence, "priority")
                            priority_et.text = priority

                        online.transport.write(ET.tostring(presence))

        if jid.bare() in self._pending:
            for item in self._pending[jid.bare()]:
                for client in self._connections.get_transport(jid):
                    client.transport.write(item)
            await self.delete_pending_presence(jid.bare())

        return None

    async def _handle_directed_presence(self, jid: JID, element: ET.Element):
        to = JID(element.attrib.get("to"))
        element.attrib["from"] = str(jid)

        if to.domain == AppConfig.app_config.host:
            for client in self._connections.get_transport(jid):
                client.transport.write(ET.tostring(element))

    async def _handle_subscribe(self, jid: JID, element: ET.Element):
        to = JID(element.attrib.get("to"))

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        # Handle local presence. Receiver client connected to the protocols
        if to.domain == AppConfig.app_config.host:
            roster = self._roster.roster_by_jid(jid)

            item = [item for item in roster
                    if ET.fromstring(item.get("item")).attrib.get("jid") == to.user
                    or ET.fromstring(item.get("item")).attrib.get("jid") == to.bare()
                    ]

            if not item:
                await self._roster.create_roster_entry(jid, to)
                roster = self._roster.roster_by_jid(jid)
                item = [item for item in roster
                        if ET.fromstring(item.get("item")).attrib["jid"] == to.user
                        or ET.fromstring(item.get("item")).attrib["jid"] == to.bare()
                        ]

            roster_item = item[0].get("item")
            et_item = ET.fromstring(roster_item)
            item_id = item[0].get("id")

            if et_item.attrib.get("ask") == "subscribe":
                return

            if et_item.attrib.get("subscription") in ["to", "both"]:
                petition = ET.Element(
                    "presence",
                    attrib={
                        "from": element.attrib.get('to'),
                        "to": element.attrib.get('from'),
                        "id": element.attrib.get('id') or str(uuid4()),
                        "type": "subscribed"
                    }
                )
                return ET.tostring(petition)

            if "ask" not in et_item.attrib:
                new_item = et_item.__copy__()
                new_item.attrib["ask"] = "subscribe"
                await self._roster.update_item(new_item, item_id)

            for client in self._connections.get_transport(to):
                petition = ET.Element(
                    "presence",
                    attrib={
                        "from": element.attrib.get('from'),
                        "to": element.attrib.get('to'),
                        "id": element.attrib.get('id') or str(uuid4()),
                        "type": "subscribe"
                    }
                )
                client.transport.write(ET.tostring(petition))

            else:
                await self._roster.store_pending_sub(to.bare(), element)

    async def _handle_subscribed(self, jid: JID, element: ET.Element):
        to = JID(element.attrib.get("to"))

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        # Handle local presence. Receiver client connected to the protocols
        if to.domain == AppConfig.app_config.host:
            roster_sender = self._roster.roster_by_jid(to)
            roster_receiver = self._roster.roster_by_jid(jid)

            if not roster_sender:
                return None

            if not roster_receiver:
                await self._roster.create_roster_entry(JID(jid.bare()), to)
                roster_receiver = self._roster.roster_by_jid(jid)

            roster_push_sender = None
            roster_push_receiver = None

            item_sender = [  # Sender appears in receiver roster
                item for item in roster_sender
                if ET.fromstring(item.get("item")).attrib.get("jid") == jid.user
            ]

            item_receiver = [  # Receiver appears in sender roster
                item for item in roster_receiver
                if ET.fromstring(item.get("item")).attrib.get("jid") == to.user
            ]

            if item_sender:
                item_id = item_sender[0].get("id")
                item_sender = item_sender[0].get("item")
                et_item_sender = ET.fromstring(item_sender)

                if et_item_sender.attrib.get("subscription") == "none":
                    new_item_sender = et_item_sender.__copy__()
                    if "ask" in new_item_sender.attrib:
                        new_item_sender.attrib.pop("ask")

                    new_item_sender.attrib["subscription"] = "to"
                    await self._roster.update_item(new_item_sender, item_id)

                    new_item_sender.attrib['jid'] = new_item_sender.attrib['jid'] + f"@{AppConfig.app_config.host}"
                    roster_push_sender = new_item_sender

                elif et_item_sender.attrib.get("subscription") == "from":
                    new_item_sender = et_item_sender.__copy__()
                    if "ask" in new_item_sender.attrib:
                        new_item_sender.attrib.pop("ask")

                    new_item_sender.attrib["subscription"] = "both"
                    await self._roster.update_item(new_item_sender, item_id)

                    new_item_sender.attrib['jid'] = new_item_sender.attrib['jid'] + f"@{AppConfig.app_config.host}"
                    roster_push_sender = new_item_sender

                else:
                    item_sender = None

            if item_receiver:
                item_id = item_receiver[0].get("id")
                item_receiver = item_receiver[0].get("item")
                et_item_receiver = ET.fromstring(item_receiver)

                if et_item_receiver.attrib.get("subscription") == "none":
                    new_item_receiver = et_item_receiver.__copy__()
                    new_item_receiver.attrib["subscription"] = "from"
                    await self._roster.update_item(new_item_receiver, item_id)

                    new_item_receiver.attrib['jid'] = new_item_receiver.attrib['jid'] + f"@{AppConfig.app_config.host}"
                    roster_push_receiver = new_item_receiver

                elif et_item_receiver.attrib["subscription"] == "to":
                    new_item_receiver = et_item_receiver.__copy__()
                    new_item_receiver.attrib["subscription"] = "both"
                    await self._roster.update_item(new_item_receiver, item_id)

                    new_item_receiver.attrib['jid'] = new_item_receiver.attrib['jid'] + f"@{AppConfig.app_config.host}"
                    roster_push_receiver = new_item_receiver

                else:
                    item_receiver = None

            if item_receiver:
                for sender in self._connections.get_transport(JID(to.bare())):
                    res = ET.Element(
                        "presence",
                        attrib={
                            "from": jid.bare(),
                            "to": str(to),
                            "id": element.attrib.get('id') or str(uuid4()),
                            "type": "subscribed",
                        },
                    )
                    sender.transport.write(ET.tostring(res))
                return None

            if roster_push_sender is not None:
                for sender in self._connections.get_transport(JID(to.bare())):
                    res = IQ(to=str(sender.jid), type_=IQ.TYPE.SET)
                    query = ET.SubElement(res, "query", attrib={"xmlns": "jabber:iq:roster"})
                    query.append(roster_push_sender)

                    sender.transport.write(ET.tostring(res))

            if roster_push_receiver is not None:
                for receiver in self._connections.get_transport(JID(jid.bare())):
                    res = IQ(to=str(receiver.jid), type_=IQ.TYPE.SET)
                    query = ET.SubElement(res, "query", attrib={"xmlns": "jabber:iq:roster"})
                    query.append(roster_push_receiver)

                    receiver.transport.write(ET.tostring(res))

    async def _handle_unsubscribed(self, jid: JID, element: ET.Element):
        to = JID(element.attrib["to"])

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        # Handle locally
        if to.domain == AppConfig.app_config.host:
            roster = self._roster.roster_by_jid(jid)
            item = [item for item in roster
                    if ET.fromstring(item.get("item")).attrib.get("jid") in [to.bare(), to.user]]

            if not item:
                return

            item_id = item[0].get("id")
            item = item[0].get("item")
            et_item = ET.fromstring(item)
            new_item = et_item.__copy__()

            updated = False

            if et_item.attrib.get("subscription") == "to":
                new_item.attrib["subscription"] = "none"
                await self._roster.update_item(item=new_item, id_=item_id)
                updated = True

            elif et_item.attrib.get("subscription") == "both":
                new_item.attrib["subscription"] = "from"
                await self._roster.update_item(item=new_item, id_=item_id)
                updated = True

            elif "ask" in et_item.attrib:
                new_item.attrib.pop("ask")
                await self._roster.update_item(item=new_item, id_=item_id)

            if updated:
                presence = ET.Element(
                    "presence",
                    attrib={
                        "from": element.attrib.get('from'),
                        "to": to.bare(),
                        "id": str(uuid4()),
                        "type": "unsubscribed"
                    }
                )
                for buffer in self._connections.get_transport(to):
                    buffer.transport.write(ET.tostring(presence))

    async def _handle_unavailable(self, jid: JID, element: ET.Element):
        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        roster = self._roster.roster_by_jid(jid)
        for item in roster:
            to = ET.fromstring(item[-1]).attrib.get('jid')

            if to.split("@")[1] == AppConfig.app_config.host:
                presence = element.__copy__()
                presence.attrib["to"] = to

                for client in self._connections.get_transport(JID(to)):
                    client.transport.write(ET.tostring(presence))




import xml.etree.ElementTree as ET
from uuid import uuid4

from pyjabber import AppConfig
from pyjabber.plugins.xep_0363.upload_server import UploadHttpServer
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class HTTPFieldUpload(metaclass=Singleton):
    __slots__ = ('_host', '_max_size', '_http_app_instance')

    def __init__(self, http_app_instance: UploadHttpServer):
        self._host = "upload.$".replace("$", AppConfig.app_config.host)
        self._max_size = AppConfig.app_config.items["upload.$"]["extra"]["max-size"]
        self._http_app_instance: UploadHttpServer = http_app_instance

    async def feed(self, jid: JID, element: ET.Element):
        if len(element) != 1:
            return SE.invalid_xml()

        request = element.find('{urn:xmpp:http:upload:0}request')
        if request is None:
            return SE.invalid_xml()

        filename = request.attrib.get('filename')
        size = request.attrib.get('size')

        try:
            size = int(size)
        except TypeError:
            return SE.not_acceptable("Missing \"size\" parameter")
        except ValueError:
            return SE.not_acceptable("Invalid \"size\" parameter. Must be an integer")

        content_type = request.attrib.get('content-type')

        if size > self._max_size:
            return SE.not_acceptable(f"File too large. The maximum file size is {self._max_size} bytes")

        slot_id = self._http_app_instance.slot_request(
            filename=filename,
            content_type=content_type,
            content_length=size
        )

        iq_res = IQ(
            type_=IQ.TYPE.RESULT,
            from_=AppConfig.app_config.host,
            to=str(jid),
            id_=element.attrib.get('id') or str(uuid4())
        )

        slot = ET.SubElement(iq_res, '{urn:xmpp:http:upload:0}slot')
        ET.SubElement(slot, '{urn:xmpp:http:upload:0}put', attrib={
            "url": f"http://{AppConfig.app_config.host}:9090/upload/{slot_id}/{filename}"
        })
        ET.SubElement(slot, '{urn:xmpp:http:upload:0}get', attrib={
            "url": f"http://{AppConfig.app_config.host}:9090/upload/{slot_id}/{filename}"
        })

        return ET.tostring(iq_res)





from loguru import logger


class TransportProxy:
    __slots__ = ('_transport', '_peer', '_server')

    def __init__(self, transport, peer, server = False):
        self._transport = transport
        self._peer = peer
        self._server = server

    @property
    def original_transport(self):
        return self._transport

    def write(self, data):
        logger.trace(f"Sending to {'protocols ' if self._server else ''}{self._peer}: {data}")
        return self._transport.write(data)

    def __getattr__(self, name):
        return getattr(self._transport, name)
